unit IdMessageClient;

interface

uses
  Classes,
  IdGlobal, IdMessage, IdTCPClient, IdHeaderList;

type
  TIdMessageClient = class(TIdTCPClient)
  protected
    // The length of the folded line
    FMsgLineLength: integer;
    // The string to be pre-pended to the next line
    FMsgLineFold: string;
    //
    procedure ReceiveBody(AMsg: TIdMessage; const ADelim: string = '.'); virtual;
    function ReceiveHeader(AMsg: TIdMessage; const ADelim: string = ''): boolean; virtual;
    procedure SendBody(AMsg: TIdMessage); virtual;
    procedure SendHeader(AMsg: TIdMessage); virtual;
    procedure WriteBodyText(AMsg: TIdMessage); virtual;
    procedure WriteFoldedLine(const ALine : string);
  public
    constructor Create(AOwner : TComponent); override;
    procedure ProcessMessage(AMsg: TIdMessage); overload;
    procedure ProcessMessage(AMsg: TIdMessage; const AStream: TStream); overload;
    procedure ProcessMessage(AMsg: TIdMessage; const AFilename: string); overload;
    procedure SendMsg(AMsg: TIdMessage); virtual;
    //
    property MsgLineLength: integer read FMsgLineLength write FMsgLineLength;
    property MsgLineFold: string read FMsgLineFold write FMsgLineFold;
  end;

implementation

uses
  IdCoder, IdCoderQuotedPrintable, IdCoder3to4, IdMessageCoderMIME, IdMessageCoderUUE, IdMEssageCoderXXE, //TODO: Remove this reference and make it completely pluggable
  IdCoderHeader, IdMessageCoder, IdComponent, IdException, IdResourceStrings, IdTCPConnection,
  IdTCPStream, IdIOHandlerStream, IdIOHandler,
  SysUtils;

function GetLongestLine(var ALine : String; ADelim : String) : String;
var
  i, fnd, lineLen, delimLen : Integer;
begin
  i := 0;
  fnd := -1;
  delimLen := length(ADelim);
  lineLen := length(ALine);
  while i < lineLen do begin
    if ALine[i] = ADelim[1] then begin
      if Copy(ALine, i, delimLen) = ADelim then begin
        fnd := i;
      end;
    end;
    Inc(i);
  end;

  if fnd = -1 then begin
    result := '';
  end else begin
    result := Copy(ALine, 1, fnd - 1);
    ALine := Copy(ALine, fnd + delimLen, lineLen);
  end;
end;

///////////////////
// TIdMessageClient
///////////////////

constructor TIdMessageClient.Create;
begin
  inherited;
  FMsgLineLength := 79;
  FMsgLineFold := #9;
end;

procedure TIdMessageClient.WriteFoldedLine;
var
  ins, s, line, spare : String;
  msgLen, insLen : Word;
begin
  s := ALine;

  // To give an amount of thread-safety
  ins := FMsgLineFold;
  insLen := length(ins);
  msgLen := FMsgLineLength;

  // Do first line
  if length(s) > FMsgLineLength then begin
    spare := Copy(s, 1, msgLen);
    line := GetLongestLine(spare, ' ');
    s := spare + Copy(s, msgLen + 1, length(s));
    WriteLn(line);

    // continue with the folded lines
    while length(s) > (msgLen - insLen) do begin
      spare := Copy(s, 1, (msgLen - insLen));
      line := GetLongestLine(spare, ' ');
      s := ins + spare + Copy(s, (msgLen - insLen) + 1, length(s));
      WriteLn(line);
    end;

    // complete the output with what's left
    if Trim(s) <> '' then begin
      WriteLn(ins + s);
    end;
  end else begin
    WriteLn(s);
  end;
end;

procedure TIdMessageClient.ReceiveBody(AMsg: TIdMessage; const ADelim: string = '.');
var
  MsgEnd: Boolean;
  LActiveDecoder: TIdMessageDecoder;
  LLine: string;

  function ProcessTextPart(ADecoder: TIdMessageDecoder): TIdMessageDecoder;
  var
    LDestStream: TStringStream;
  begin
    LDestStream := TStringStream.Create(''); try
      Result := ADecoder.ReadBody(LDestStream, MsgEnd);
      ADecoder.Free;
      with TIdText.Create(AMsg.MessageParts) do begin
        Body.Text := LDestStream.DataString;
      end;
    finally FreeAndNil(LDestStream); end;
  end;

  function ProcessAttachment(ADecoder: TIdMessageDecoder): TIdMessageDecoder;
  var
    LDestStream: TFileStream;
    LTempPathname: string;
  begin
    LTempPathname := MakeTempFilename;
    LDestStream := TFileStream.Create(LTempPathname, fmCreate); try
      Result := ADecoder.ReadBody(LDestStream, MsgEnd);
      with TIdAttachment.Create(AMsg.MessageParts) do begin
        Filename := ADecoder.Filename;
        StoredPathname := LTempPathname;
      end;
      ADecoder.Free;
    finally FreeAndNil(LDestStream); end;
  end;

begin
  MsgEnd := False;
  if AMsg.NoDecode then begin
    Capture(AMsg.Body, ADelim);
  end else begin
    BeginWork(wmRead); try
      LActiveDecoder := nil;
      repeat
        LLine := ReadLn;
        if LLine = ADelim then begin
          Break;
        end;
        if LActiveDecoder = nil then begin
          LActiveDecoder := TIdMessageDecoderList.CheckForStart(AMsg, LLine);
        end;
        if LActiveDecoder = nil then begin
          AMsg.Body.Add(LLine);
        end else begin
          while LActiveDecoder <> nil do begin
            LActiveDecoder.SourceStream := TIdTCPStream.Create(Self);
            LActiveDecoder.ReadHeader;
            case LActiveDecoder.PartType of
              ptUnknown: begin
                raise EIdException.Create(RSMsgClientUnkownMessagePartType);
              end;
              ptText: begin
                LActiveDecoder := ProcessTextPart(LActiveDecoder);
              end;
              ptAttachment: begin
                LActiveDecoder := ProcessAttachment(LActiveDecoder);
              end;
            end;
          end;
        end;
      until MsgEnd;
    finally
      EndWork(wmRead);
    end;
  end;
end;



procedure TIdMessageClient.SendHeader(AMsg: TIdMessage);
var
  LHeaders: TIdHeaderList;
begin
  LHeaders := AMsg.GenerateHeader; try
    WriteStrings(LHeaders);
  finally FreeAndNil(LHeaders); end;
end;

procedure TIdMessageClient.SendBody(AMsg: TIdMEssage);
var
  i: Integer;
  LAttachment: TIdAttachment;
  LBoundary: string;
  LDestStream: TIdTCPStream;
  LMIMEAttachments: boolean;
  ISOCharset: string;
  HeaderEncoding: Char;  { B | Q }
  TransferEncoding: TTransfer;

  procedure WriteTextPart(ATextPart: TIdText);
  var
    Data: string;
    i: Integer;
  begin
    if Length(ATextPart.ContentType) = 0 then
      ATextPart.ContentType := 'text/plain'; {do not localize}
    if Length(ATextPart.ContentTransfer) = 0 then
      ATextPart.ContentTransfer := 'quoted-printable'; {do not localize}

    WriteLn('Content-Type: ' + ATextPart.ContentType); {do not localize}
    WriteLn('Content-Transfer-Encoding: ' + ATextPart.ContentTransfer); {do not localize}
    WriteStrings(ATextPart.ExtraHeaders);
    WriteLn('');
    // TODO: Provide B64 encoding later
//    if AnsiSameText(ATextPart.ContentTransfer, 'base64') then begin
//      LEncoder := TIdEncoder3to4.Create(nil);
    if AnsiSameText(ATextPart.ContentTransfer, 'quoted-printable') then begin
      for i := 0 to ATextPart.Body.Count - 1 do
      begin
        if Copy(ATextPart.Body[i], 1, 1) = '.' then begin
          ATextPart.Body[i] := '.' + ATextPart.Body[i];
        end;

        Data := TIdEncoderQuotedPrintable.EncodeString(ATextPart.Body[i] + EOL);
        if TransferEncoding = iso2022jp then
          Write(Encode2022JP(Data))
        else
          Write(Data);
      end;
    end else begin
      WriteStrings(ATextPart.Body);
    end;
    WriteLn('');
  end;

begin
  LMIMEAttachments := AMsg.Encoding = meMIME;
  LBoundary := '';
  InitializeISO(TransferEncoding, HeaderEncoding, ISOCharSet);
  BeginWork(wmWrite); try
    if AMsg.MessageParts.AttachmentCount > 0 then begin
      if LMIMEAttachments then begin
        WriteLn('This is a multi-part message in MIME format'); {do not localize}
        WriteLn('');
        if AMsg.MessageParts.RelatedPartCount > 0 then begin
          LBoundary := IndyMultiPartRelatedBoundary;
        end else begin
          LBoundary := IndyMIMEBoundary;
        end;
        WriteLn('--' + LBoundary);
      end else begin
        // It's UU, write the body
        WriteBodyText(AMsg);
        WriteLn('');
      end;
      if AMsg.MessageParts.TextPartCount > 1 then begin
        WriteLn('Content-Type: multipart/alternative; '); {do not localize}
        WriteLn('        boundary="' + IndyMultiPartAlternativeBoundary + '"'); {do not localize}
        WriteLn('');
        for i := 0 to AMsg.MessageParts.Count - 1 do begin
          if AMsg.MessageParts.Items[i] is TIdText then begin
            WriteLn('--' + IndyMultiPartAlternativeBoundary);
            DoStatus(hsStatusText,  [RSMsgClientEncodingText]);
            WriteTextPart(AMsg.MessageParts.Items[i] as TIdText);
            WriteLn('');
          end;
        end;
        WriteLn('--' + IndyMultiPartAlternativeBoundary + '--');
      end else begin
        if LMIMEAttachments then begin
          WriteLn('Content-Type: text/plain'); {do not localize}
          WriteLn('Content-Transfer-Encoding: 7bit'); {do not localize}
          WriteLn('');
          WriteBodyText(AMsg);
        end;
      end;
      // Send the attachments
      for i := 0 to AMsg.MessageParts.Count - 1 do begin
        if AMsg.MessageParts[i] is TIdAttachment then begin
          LAttachment := TIdAttachment(AMsg.MessageParts[i]);
          DoStatus(hsStatusText, [RSMsgClientEncodingAttachment]);
          if LMIMEAttachments then begin
            WriteLn('');
            WriteLn('--' + LBoundary);
            if Length(LAttachment.ContentTransfer) = 0 then begin
              LAttachment.ContentTransfer := 'base64'; {do not localize}
            end;
            if Length(LAttachment.ContentDisposition) = 0 then begin
              LAttachment.ContentDisposition := 'attachment'; {do not localize}
            end;
            if (LAttachment.ContentTransfer = 'base64') {do not localize}
             and (Length(LAttachment.ContentType) = 0) then begin
              LAttachment.ContentType := 'application/octet-stream'; {do not localize}
            end;
            WriteLn('Content-Type: ' + LAttachment.ContentType + ';'); {do not localize}
            WriteLn('        name="' + ExtractFileName(LAttachment.FileName) + '"'); {do not localize}
            WriteLn('Content-Transfer-Encoding: ' + LAttachment.ContentTransfer); {do not localize}
            WriteLn('Content-Disposition: ' + LAttachment.ContentDisposition +';'); {do not localize}
            WriteLn('        filename="' + ExtractFileName(LAttachment.FileName) + '"'); {do not localize}
            WriteStrings(LAttachment.ExtraHeaders);
            WriteLn('');
          end;
          LDestStream := TIdTCPStream.Create(Self); try
            TIdAttachment(AMsg.MessageParts[i]).Encode(LDestStream);
          finally FreeAndNil(LDestStream); end;
          WriteLn('');
        end;
      end;
      if LMIMEAttachments then begin
        WriteLn('--' + LBoundary + '--');
      end;
    end else if AMsg.MessageParts.TextPartCount > 1 then begin
      WriteLn('This is a multi-part message in MIME format'); {do not localize}
      WriteLn('');
      for i := 0 to AMsg.MessageParts.Count - 1 do
      begin
        if AMsg.MessageParts.Items[i] is TIdText then begin
          WriteLn('--' + IndyMIMEBoundary);
          DoStatus(hsStatusText, [RSMsgClientEncodingText]);
          WriteTextPart(AMsg.MessageParts.Items[i] as TIdText);
        end;
      end;
      WriteLn('--' + IndyMIMEBoundary + '--');
    end else begin
      DoStatus(hsStatusText, [RSMsgClientEncodingText]);
      // Write out Body
      //TODO: Why just iso2022jp? Why not someting generic for all MBCS? Or is iso2022jp special?
      if TransferEncoding = iso2022jp then begin
        for i := 0 to AMsg.Body.Count - 1 do begin
          if Copy(AMsg.Body[i], 1, 1) = '.' then begin
            WriteLn('.' + Encode2022JP(AMsg.Body[i]));
          end else begin
            WriteLn(Encode2022JP(AMsg.Body[i]));
          end;
        end;
      end else begin
        WriteBodyText(AMsg);
      end;
    end;
  finally EndWork(wmWrite); end;
end;

procedure TIdMessageClient.SendMsg(AMsg: TIdMessage);
begin
  SendHeader(AMsg);
  WriteLn('');
  SendBody(AMsg);
end;

// S.G. 27.06.00: Now returns true if the message contains additional parts and false if the message contains only headers
function TIdMessageClient.ReceiveHeader(AMsg: TIdMessage; const ADelim: string = ''): boolean;
var
  NewLine: String;
begin
//  Capture(AMsg.Headers, ADelim, True);
  Result := true;
  BeginWork(wmRead);
  try
    repeat
      NewLine := ReadLn;
      if NewLine = ADelim then begin
        break;
      end else if Copy(NewLine, 1, 2) = '..' then
        NewLine := '.'
      else if newLine = '.' then
      begin
        result := false;
        break;
      end;
      AMsg.Headers.Append(NewLine);
    until False;
    AMsg.ProcessHeaders;
  finally
    EndWork(wmRead);
  end
end;

procedure TIdMessageclient.ProcessMessage(AMsg: TIdMessage);
begin
  if IOHandler <> NIL Then Begin
    ReceiveHeader(AMsg);
    ReceiveBody(AMsg);
  end;
end;

procedure TIdMessageClient.ProcessMessage(AMsg: TIdMessage; const AStream: TStream);
var
  LVoidStream: TMemoryStream;
  LVoidIOHandler: TIdIOHandler;
begin
  LVoidStream := TMemoryStream.Create; try
    LVoidIOHandler := IOHandler;
    IOHandler := TIdIOHandlerStream.Create(nil); try
    TIdIOHandlerStream(IOHandler).ReadStream := AStream;
    TIdIOHandlerStream(IOHandler).WriteStream := LVoidStream;
    ReceiveHeader(AMsg);
    ReceiveBody(AMsg);
    finally
      IOHandler.Free;
      IOHandler := LVoidIOHandler;
    end;
  finally FreeAndNil(LVoidStream); end;
end;

procedure TIdMessageClient.ProcessMessage(AMsg: TIdMessage; const AFilename: string);
var
  LStream: TFileStream;
begin
  LStream := TFileStream.Create(AFileName, fmOpenRead); try
    ProcessMessage(AMsg, LStream);
  finally FreeAndNil(LStream); end;
end;

procedure TIdMessageClient.WriteBodyText(AMsg: TIdMessage);
var
  i: integer;
begin
  for i := 0 to AMsg.Body.Count - 1 do begin
    if Copy(AMsg.Body[i], 1, 1) = '.' then begin
      WriteLn('.' + AMsg.Body[i]);
    end else begin
      WriteLn(AMsg.Body[i]);
    end;
  end;
end;

end.
